/*{{{}}}*/
/*{{{  #include*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#if (defined(__unix__) || defined(unix)) && !defined(USG)
    #include <sys/param.h>
#endif
#ifdef MSDOS
    #include <string.h>
    #include <time.h>
#else
    #ifdef BSD
        #include <strings.h>
        #include <time.h>
    #else
        #include <string.h>
        #include <time.h>
    #endif
#endif
#include "fft.h"
#include "calcroll.h"
/*}}}  */
/*{{{  global variables*/
char *myname = "calcroll";
char *help = "Calculate FIR rolloff filter coeffiecients";
char *version = "ver. 2.7.2 (Oct. 28, 2021)";
char *copyright = "(c) 2000 Takayuki HOSODA, Finetune co., ltd.";
char *disclaimer = "calcroll is free software with ABSOLUTELY NO WARRANTY.";
char *usage = "\
SYNOPSIS\n\
    calcroll [OPTIONS] [FILENAME]\n\
OPTIONS\n\
    -cdouble     : format output for floating C program\n\
    -cint        : format output for integer C program\n\
    -distrib=f.f : distribution rate (0 < rolloff <= 1.0)\n\
                   default=0.5 (root cosine rolloff)\n\
    -fs=nn       : over sampling (fs = 2^N) default=8\n\
    -help        : show this help\n\
    -rolloff=f.f : rolloff rate (0 <= rolloff <= 1.0)\n\
                   default=0.5 (typical for most radio communications)\n\
    -scale=f.f   : scaling factor (0 < scale) default=1.0\n\
    -symmetry    : symmetrical time domain output\n\
    -taps=nn     : taps FIR filter (taps = 2^N) default=64\n\
    -verbose     : show verbose informations\n\
EXAMPLE\n\
    calcroll -fs=4 -taps=32 -rolloff=0.5 -distrib=0.5 -scale=1024 rolloff.dat\n\
BUGS\n\
    Bug reports, comments, suggestions, are always walcome,\n\
    please contact us at http://www.finetune.co.jp or E-mail to\n\
    postmaster@finetune.co.jp\n\
";
/*}}}  */
/*{{{  errmsg(err)*/
int
errmsg(err)
    int err;
{
    switch(err){
    case NO_ERR:
        break;
    case E_HELP:
        fprintf(stdout, "%s: %s %s\n%s%s", myname, help, version, copyright, usage);
        break;
    case E_OPTION:
        fprintf(stderr, "%s: Incorrect option.\n", myname);
        break;
    case E_FOPEN:
        fprintf(stderr, "%s: Cannot open output file.\n", myname);
        break;
    case E_MALLOC:
        fprintf(stderr, "%s: Not enough memory.\n", myname);
        break;
    case E_FS:
        fprintf(stderr, "%s: Incorrect number of fs.\n", myname);
        break;
    case E_TAPS:
        fprintf(stderr, "%s: Incorrect number of taps.\n", myname);
        break;
    case E_ROLLOFF:
        fprintf(stderr, "%s: rolloff rate is out of range.\n", myname);
        break;
    case E_DISTRIB:
        fprintf(stderr, "%s: distribution rate is out of range.\n", myname);
        break;
    default:
        fprintf(stderr, "%s: Unknown error occured.\n", myname);
        break;
    }
    return(err);
}
/*}}}  */
/*{{{  parse(argc, argv, flags, fp, fs, taps, rolloff, distrib, scale)*/
int
parse(argc, argv, flags, fp, fs, taps, rolloff, distrib, scale)
    /*{{{  variables*/
    int             argc;
    char           *argv[];
    int            *flags;
    FILE          **fp;
    int            *fs;
    int            *taps;
    double         *rolloff;   /* rolloff rate */
    double         *distrib;   /* distribution rate */
    double         *scale;
    /*}}}  */
{
    /*{{{  variables*/
    int             i;
    int          rtnv;
    int           tmp;
    /*}}}  */
    /*{{{  initialise*/
    rtnv = 0;
    *flags = 0;
    *fp = stdout;
    *rolloff = 0.5;
    *distrib = 0.5; /* root rolloff */
    *taps = 64;
    *fs = 8;
    *scale = 1.0;
    /*}}}  */
    /*{{{  get options*/
    for (i = 1; i < argc; i++) {
        if (*argv[i] == '-') {
            /*{{{  options*/
            switch (*++argv[i]) {
            /*{{{  cdbl, cint*/
            case 'c':
                if (!strncmp(argv[i], "cdouble", strlen("cd"))) {
                    *flags |= F_C_DBL;
                } else if(!strncmp(argv[i], "cint", strlen("ci"))){
                    *flags |= F_C_INT;
                }else{
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  distrib*/
            case 'd':
                if (!strncmp(argv[i], "distrib=", strlen("distrib="))) {
                    sscanf(&argv[i][strlen("distrib=")], "%lg", distrib);
                } else {
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  fs*/
            case 'f':
                if (!strncmp(argv[i], "fs=", strlen("fs="))) {
                    sscanf(&argv[i][strlen("fs=")], "%d", fs);
                } else {
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  help*/
            case 'h':
                rtnv = E_HELP;
                break;
            /*}}}  */
            /*{{{  rate*/
            case 'r':
                if (!strncmp(argv[i], "rolloff=", strlen("rolloff="))) {
                    sscanf(&argv[i][strlen("rolloff=")], "%lg", rolloff);
                } else {
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  scale*/
            case 's':
                if (!strncmp(argv[i], "scale=", strlen("scale="))) {
                    sscanf(&argv[i][strlen("scale=")], "%lg", scale);
                } else if(!strncmp(argv[i], "symmetry", strlen("sym"))) {
                    *flags |= F_SYM;
                } else {
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  taps*/
            case 't':
                if (!strncmp(argv[i], "taps=", strlen("taps="))) {
                    sscanf(&argv[i][strlen("taps=")], "%d", taps);
                } else {
                    rtnv = E_OPTION;
                }
                break;
            /*}}}  */
            /*{{{  verbose*/
            case 'v':
                *flags |= F_VERBOSE;
                break;
            /*}}}  */
            /*{{{  default*/
            default:
                rtnv = E_OPTION;
                break;
            /*}}}  */
            }
            --argv[i];
            /*}}}  */
        }else{
            if(NULL == (*fp = fopen(argv[i], "w"))){
                rtnv = E_FOPEN;
            }
        }
    }
    /*}}}  */
    /*{{{  check value*/
    /*{{{  check rolloff rate*/
    if((*rolloff < 0.0) || (*rolloff > 1.0)){
        *rolloff = 0.0;
        rtnv = E_ROLLOFF;
    }
    /*}}}  */
    /*{{{  check distribution rate*/
    if((*distrib < 0.0) || (*distrib > 1.0)){
        *distrib = 0.5;
        rtnv = E_DISTRIB;
    }
    /*}}}  */
    /*{{{  check scaling factor*/
    if(*scale <= 0.0){
        *scale = 1.0;
        rtnv = E_SCALE;
    }
    /*}}}  */
    /*{{{  check fs*/
    if(*fs <= 0){
        *fs = 4;
        rtnv = E_FS;
    }else{
        /*{{{  is fs N-th powers of 2*/
        tmp = 0;
        for(i = 0; i < sizeof(int)*8; i++){
            tmp += ((*fs >> i) & 1);
        }
        if(tmp != 1){
            *fs = 4;
            rtnv = E_FS;
        }
        /*}}}  */
    }
    /*}}}  */
    /*{{{  check taps*/
    if(*taps <= 1){
        *taps = 4;
        rtnv = E_FS;
    }else{
        /*{{{  is taps N-th powers of 2*/
        tmp = 0;
        for(i = 0; i < sizeof(int)*8; i++){
            tmp += ((*taps >> i) & 1);
        }
        if(tmp != 1){
            *taps = 32;
            rtnv = E_TAPS;
        }
        /*}}}  */
    }
    /*}}}  */
    /*}}}  */
    return (rtnv);
}
/*}}}  */
/*{{{  main()*/
int
main(argc, argv)
    int argc;
    char *argv[];
{
    /*{{{  variables*/
    int       n;  /* taps */
    int      fs;  /* fs-th over sampling */
    double    a;  /* roll off rate */
    double    d;  /* distribution rate */
    double    t;  /* temporary */
    double    s;  /* scaling factor */
    double  mag;  /* magnitude */
    double   ph;  /* phase */
    
    int       (*freq)();  /* pointer to frequency response function */
    double    *buf;
    double    *re;
    double    *im;
    fft_plan  *p;
    
    int i;
    FILE      *outfile;
    int       flags = 0;
    int       rtnv = 0;
    time_t    tloc;
    
    char *c_comment[2] = {"/*", "*/"};
    char *sh_comment[2] = {"#", " "};
    char *comment_begin;
    char *comment_end;
    int    sum=0;
    double sumd=0.0;
    double delay=0.0;
    /*}}}  */
    /*{{{  initailize variables*/
    freq = rolloff;
    /*}}}  */
    /*{{{  parse*/
    rtnv = parse(argc, argv, &flags, &outfile, &fs, &n, &a, &d, &s);
    /*}}}  */
    /*{{{  initialize variables after parse*/
    if((flags & F_C_DBL) || (flags & F_C_INT)){
        comment_begin = c_comment[0];
        comment_end = c_comment[1];
    }else{
        comment_begin = sh_comment[0];
        comment_end = sh_comment[1];
    }
    /*}}}  */
    if(!rtnv){
        /*{{{  allocate memory*/
        if(NULL == (buf = (double *)malloc(4*n * sizeof(double)))){
            rtnv = E_MALLOC;
        }
        /*}}}  */
        if(!rtnv){
            /*{{{  initialize buffers*/
            re = buf + (2 * n);
            im = buf + (3 * n);
            for(i = 4*n-1; i >= 0; i--){
                buf[i] = 0.0;
            }
            /*}}}  */
            /*{{{  set delay*/
            if(flags & F_SYM){
                delay = -M_PI * (double)(n-1)/(double)n;
            }else if((flags & F_C_INT) || (flags & F_C_DBL)){
                delay = -M_PI * (double)(n-2)/(double)n;
            }else{
                delay = -M_PI;
            }
            /*}}}  */
            /*{{{  set frequency response*/
            for(i = 0; i <= n/2; i++){
                t = (double)(i * fs) / ((double)n);
                rolloff(t, a, d, &mag, &ph);
                ph += delay * (double)i;   /* add constant delay*/
                re[i] = mag * cos(ph);
                im[i] = mag * sin(ph);
            }
            for(i = n/2; i < n; i++){
                re[i] =  re[n-i];
                im[i] = -im[n-i];
            }
            /*{{{  save re[], im[] to buf[]*/
            for(i = 0; i < (n + n); i++){
                buf[i] = re[i];
            }
            /*}}}  */
            /*}}}  */
            /*{{{  create reverse FFT plan*/
            p = pre_fft(n, FFT_BACKWARD);
            /*}}}  */
            /*{{{  do one dimensional FFT*/
            fft(p, re, im);
            /*}}}  */
            /*{{{  output header*/
            if(flags & F_VERBOSE){
                time(&tloc);
                fprintf(outfile, "%s %s: %s\n * %s\n * %s * ",
                    comment_begin,
                    myname, version,
                    copyright,
                    ctime(&tloc)
                    );
            } else {
                fprintf(outfile, "%s ", comment_begin);
            }
            fprintf(outfile, "fs=%d, taps=%d, rolloff=%f, distrib=%f scale=%f %s\n",
                fs, n, a, d, s,
                comment_end
                );
            /*}}}  */
            /*{{{  output results*/
            t = 0.0;
            for(i = 0; i < n; i++){
                t += re[i];
            }
            s /= t;
            if(flags & F_C_INT){
                /*{{{  C prog. interger ver.*/
                n = (flags & F_SYM) ? n : n-1;
                fprintf(outfile, "#define FSAMPLE %d\n", fs);
                fprintf(outfile, "#define NUMTAP %d\n", n);
                fprintf(outfile, "#define ROLLOFF %f\n", a);
                fprintf(outfile, "#define DISTRIB %f\n", d);
                fprintf(outfile, "#include \"rolloff.h\"\n");
                fprintf(outfile, "extern int rollbuf[NUMTAP];\n");
                fprintf(outfile, "int rollref[NUMTAP] = {\\\n    ");
                for(i = 0; i < n; i++){
                    sum += (int)s*re[i];
                    fprintf(outfile, "%d%s", (int)(s*re[i]),
                        i == (n-1) ? "};\n" : (i % 8) == 7 ? ", \\\n    " : ", ");
                }
                fprintf(outfile, "#define DIVROLL %d\n", sum);
                if(sum < n){
                    fprintf(stderr, "%s: WARNING! DIVROLL is too small. You'd better check the scaling factor.\n",
                        myname);
                }
                /*}}}  */
            }else if(flags & F_C_DBL){
                /*{{{  C prog. double ver.*/
                n = (flags & F_SYM) ? n : n-1;
                fprintf(outfile, "#define FSAMPLE %d\n", fs);
                fprintf(outfile, "#define NUMTAP %d\n", n);
                fprintf(outfile, "#define ROLLOFF %f\n", a);
                fprintf(outfile, "#define DISTRIB %f\n", d);
                fprintf(outfile, "#include \"rolloff.h\"\n");
                fprintf(outfile, "extern double rollref[NUMTAP];\n");
                fprintf(outfile, "double rollref[NUMTAP] = {\\\n    ");
                for(i = 0; i < n; i++){
                    sumd += s*re[i];
                    fprintf(outfile, "%e%s", (s*re[i]),
                        i == (n-1) ? "};\n" : (i % 4) == 3 ? ", \\\n    " : ", ");
                }
                fprintf(outfile, "#define DIVROLL %e\n", sumd);
                /*}}}  */
            }else if(flags & F_VERBOSE){
                /*{{{  n, time, freq.re, freq.im*/
                fprintf(outfile, "%s N  time  freq.re  freq.im%s\n",
                    comment_begin, comment_end);
                for(i = 0; i <  n; i++){
                    fprintf(outfile, "%d %f %e %e\n", i, s*re[i], buf[i], buf[n+i]);
                }
                /*}}}  */
            }else{
                /*{{{  time*/
                for(i = 0; i < n; i++){
                    fprintf(outfile, "%f\n", s*re[i]);
                }
                /*}}}  */
            }
            /*}}}  */
            /*{{{  destroy FFT plan*/
            post_fft();
            /*}}}  */
            /*{{{  free memory*/
            free(buf);
            /*}}}  */
        }
    }
    return(errmsg(rtnv));
}
/*}}}  */
